/*
 * Work under Copyright. Licensed under the EUPL.
 * See the project README.md and LICENSE.txt for more information.
 */

package net.dries007.tfc.world.classic.worldgen.trees;

import java.util.Random;

import net.minecraft.block.BlockLog;
import net.minecraft.util.ResourceLocation;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.World;
import net.minecraft.world.gen.structure.template.PlacementSettings;
import net.minecraft.world.gen.structure.template.Template;
import net.minecraft.world.gen.structure.template.TemplateManager;

import net.dries007.tfc.TerraFirmaCraft;
import net.dries007.tfc.api.ITreeGenerator;
import net.dries007.tfc.api.types.Tree;
import net.dries007.tfc.api.util.TFCConstants;
import net.dries007.tfc.objects.blocks.wood.BlockLeavesTFC;
import net.dries007.tfc.objects.blocks.wood.BlockLogTFC;

import static net.dries007.tfc.objects.blocks.wood.BlockLogTFC.PLACED;
import static net.minecraft.block.BlockLog.LOG_AXIS;

/**
 * This is a tree generator only used for the willow tree shapes. Requires two structure blocks:
 * both found in /assets/tfc/[TREE NAME]/, named base.nbt and overlay.nbt respectively. See the examples for
 * TFC willow tree for what the structure blocks should look like.
 */
public class TreeGenWillow implements ITreeGenerator
{
    private static final PlacementSettings settingsFull = ITreeGenerator.getDefaultSettings();
    private static final PlacementSettings settingsWeak = ITreeGenerator.getDefaultSettings().func_189946_a(0.5F);
    private Template structureBase;
    private Template structureOverlay;

    @Override
    public void generateTree(TemplateManager manager, World world, BlockPos pos, Tree tree, Random rand)
    {
        ResourceLocation base = new ResourceLocation(TFCConstants.MOD_ID, tree.getRegistryName().func_110623_a() + "/base");
        ResourceLocation overlay = new ResourceLocation(TFCConstants.MOD_ID, tree.getRegistryName().func_110623_a() + "/overlay");

        structureBase = manager.func_189942_b(world.func_73046_m(), base);
        structureOverlay = manager.func_189942_b(world.func_73046_m(), overlay);

        if (structureBase == null || structureOverlay == null)
        {
            TerraFirmaCraft.getLog().warn("Unable to find a template for " + base.toString() + " or " + overlay.toString());
            return;
        }

        int height = 5 + rand.nextInt(3),
            branches = 2 + rand.nextInt(3), x1, z1, y1;
        for (int n = 0; n <= height; n++)
        {
            if (n > 3)
                createLeafGroup(world, pos.func_177981_b(n));
            tryPlaceLog(world, pos.func_177981_b(n), tree, BlockLog.EnumAxis.Y);
        }

        for (int n = 0; n < branches; n++)
        {
            x1 = (rand.nextBoolean() ? 1 : -1) * (1 + rand.nextInt(3));
            z1 = (rand.nextBoolean() ? 1 : -1) * (1 + rand.nextInt(3));
            y1 = 3 + rand.nextInt(2);
            createLeafGroup(world, pos.func_177982_a(x1, y1, z1));
            createBranch(world, pos, x1, y1, z1, rand, tree);
        }
    }

    private void createBranch(World world, BlockPos pos1, int x, int y, int z, Random rand, Tree tree)
    {
        int x1 = x < 0 ? 1 : -1,
            z1 = z < 0 ? 1 : -1;
        do
        {
            if (x != 0 && rand.nextBoolean())
                x += x1;
            if (z != 0 && rand.nextBoolean())
                z += z1;
            tryPlaceLog(world, pos1.func_177982_a(x, y, z), tree, BlockLog.EnumAxis.NONE);
            if (rand.nextBoolean())
                createLeafGroup(world, pos1.func_177982_a(x, y, z));
        }
        while (Math.abs(x) + Math.abs(z) > 0);
    }

    private void createLeafGroup(World world, BlockPos pos)
    {
        BlockPos size = structureBase.func_186259_a();
        pos = pos.func_177982_a(-size.func_177958_n() / 2, -size.func_177956_o() / 2, -size.func_177952_p() / 2);

        ITreeGenerator.addStructureToWorld(world, pos, structureBase, settingsFull);
        ITreeGenerator.addStructureToWorld(world, pos, structureOverlay, settingsWeak);
    }

    private void tryPlaceLog(World world, BlockPos pos, Tree tree, BlockLog.EnumAxis axis)
    {
        if (world.func_180495_p(pos).func_185904_a().func_76222_j() || world.func_180495_p(pos).func_177230_c() instanceof BlockLeavesTFC)
            world.func_175656_a(pos, BlockLogTFC.get(tree).func_176223_P().func_177226_a(field_176299_a, axis).func_177226_a(PLACED, false));
    }
}
